from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, TypeVar

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal

from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers import _legend, _text_utils
from whitecanvas.layers._base import (
    HoverableDataBoundLayer,
    LayerEvents,
    PrimitiveLayer,
)
from whitecanvas.layers._mixin import EnumArray
from whitecanvas.layers._primitive.text import Texts
from whitecanvas.layers._sizehint import xy_size_hint
from whitecanvas.protocols import LineProtocol, MultiLineProtocol
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    OrientationLike,
    StepStyle,
    Symbol,
    XYData,
    _Void,
)
from whitecanvas.utils.normalize import (
    arr_color,
    as_array_1d,
    as_color_array,
    normalize_xy,
    parse_texts,
)
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers import group as _lg

_void = _Void()
_Line = TypeVar("_Line", bound=LineProtocol)


class LineLayerEvents(LayerEvents):
    color = Signal(np.ndarray)
    width = Signal(float)
    style = Signal(str)
    antialias = Signal(bool)
    clicked = Signal(int)


class LineMixin(PrimitiveLayer[_Line]):
    events: LineLayerEvents
    _events_class = LineLayerEvents

    @property
    def color(self) -> NDArray[np.floating]:
        """Color of the line."""
        return self._backend._plt_get_edge_color()

    @color.setter
    def color(self, color: ColorType):
        col = arr_color(color)
        self._backend._plt_set_edge_color(col)
        self.events.color.emit(col)

    @property
    def width(self) -> float:
        """Width of the line."""
        return self._backend._plt_get_edge_width()

    @width.setter
    def width(self, width: float):
        if not is_real_number(width):
            raise TypeError(f"Width must be a number, got {type(width)}")
        if width < 0:
            raise ValueError(f"Width must be non-negative, got {width!r}")
        w = float(width)
        self._backend._plt_set_edge_width(w)
        self.events.width.emit(w)

    @property
    def style(self) -> LineStyle:
        """Style of the line."""
        return LineStyle(self._backend._plt_get_edge_style())

    @style.setter
    def style(self, style: str | LineStyle):
        s = LineStyle(style)
        self._backend._plt_set_edge_style(s)
        self.events.style.emit(s.value)

    @property
    def alpha(self) -> float:
        return float(self.color[3])

    @alpha.setter
    def alpha(self, value: float):
        if not 0 <= value <= 1:
            raise ValueError(f"Alpha must be between 0 and 1, got {value!r}")
        self.color = (*self.color[:3], value)

    @property
    def antialias(self) -> bool:
        """Whether to use antialiasing."""
        return self._backend._plt_get_antialias()

    @antialias.setter
    def antialias(self, antialias: bool) -> None:
        if not isinstance(antialias, bool):
            raise TypeError(f"Expected antialias to be bool, got {type(antialias)}")
        self._backend._plt_set_antialias(antialias)
        self.events.antialias.emit(antialias)

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        alpha: float | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if width is not _void:
            self.width = width
        if style is not _void:
            self.style = style
        if alpha is not _void:
            self.alpha = alpha
        if antialias is not _void:
            self.antialias = antialias
        return self

    def _as_legend_item(self) -> _legend.LineLegendItem:
        return _legend.LineLegendItem(self.color, self.width, self.style)


class _SingleLine(
    LineMixin[LineProtocol], HoverableDataBoundLayer[LineProtocol, XYData]
):
    _backend_class_name = "MonoLine"
    events: LineLayerEvents
    _events_class = LineLayerEvents

    def _norm_layer_data(self, data: Any) -> XYData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 2:
                raise ValueError(f"Expected data to be (N, 2), got {data.shape}")
            xdata, ydata = data[:, 0], data[:, 1]
        else:
            xdata, ydata = data
            if xdata is None:
                xdata = self.data.x
            else:
                xdata = as_array_1d(xdata)
            if ydata is None:
                ydata = self.data.y
            else:
                ydata = as_array_1d(ydata)
        if xdata.size != ydata.size:
            raise ValueError(
                "Expected xdata and ydata to have the same size, "
                f"got {xdata.size} and {ydata.size}"
            )
        return XYData(xdata, ydata)

    def set_data(
        self,
        xdata: ArrayLike1D | None = None,
        ydata: ArrayLike1D | None = None,
    ):
        """Set x and y data of the line."""
        self.data = xdata, ydata

    @property
    def ndata(self) -> int:
        """Number of data points."""
        return self.data.x.size

    def _data_to_backend_data(self, data: XYData) -> XYData:
        return data

    def with_markers(
        self,
        symbol: Symbol | str = Symbol.CIRCLE,
        size: float | None = None,
        color: ColorType | _Void = _void,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _lg.Plot[Self]:
        """
        Add markers at each data point.

        >>> canvas.add_line(x, y).with_markers(color="yellow")

        Parameters
        ----------
        symbol : str or Symbol, default Symbol.CIRCLE
            Marker symbols.
        size : float, optional
            Marker size. Use theme default if not given.
        color : color-like, optional
            Marker face colors. Set to the line color by default.
        alpha : float, default 1.0
            The alpha channel value of the markers.
        hatch : str or FacePattern, default FacePattern.SOLID
            The marker face hatch.

        Returns
        -------
        Plot
            The plot layer.
        """
        from whitecanvas.layers._primitive import Markers
        from whitecanvas.layers.group import Plot

        if color is _void:
            color = self.color
        size = theme._default("markers.size", size)
        markers = Markers(
            *self.data, symbol=symbol, size=size, color=color, alpha=alpha,
            hatch=hatch, name=f"markers-of-{self.name}", backend=self._backend_name,
        )  # fmt: skip
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return Plot(self, markers, name=old_name)

    def with_xerr(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float = 0,
    ) -> _lg.LabeledLine[Self]:
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group import LabeledLine

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.color
        if width is _void:
            width = self.width
        if style is _void:
            style = self.style
        if antialias is _void:
            antialias = self.antialias
        xerr = Errorbars(
            self.data.y, self.data.x - err, self.data.x + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            name=f"xerr-of-{self.name}", orient=Orientation.HORIZONTAL,
            backend=self._backend_name
        )  # fmt: skip
        yerr = Errorbars.empty_v(f"yerr-of-{self.name}", backend=self._backend_name)
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LabeledLine(self, xerr, yerr, name=old_name)

    def with_yerr(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float = 0,
    ) -> _lg.LabeledLine[Self]:
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group import LabeledLine

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.color
        if width is _void:
            width = self.width
        if style is _void:
            style = self.style
        if antialias is _void:
            antialias = self.antialias
        yerr = Errorbars(
            self.data.x, self.data.y - err, self.data.y + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            name=f"yerr-of-{self.name}", backend=self._backend_name
        )  # fmt: skip
        xerr = Errorbars.empty_h(f"xerr-of-{self.name}", backend=self._backend_name)
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LabeledLine(self, xerr, yerr, name=old_name)

    def with_xband(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        *,
        color: ColorType | _Void = _void,
        alpha: float = 0.3,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _lg.LineBand[Self]:
        from whitecanvas.layers._primitive import Band
        from whitecanvas.layers.group import LineBand

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.color
        data = self.data
        _x, _y0 = self._data_to_backend_data(XYData(data.y, data.x - err))
        _, _y1 = self._data_to_backend_data(XYData(data.y, data.x + err_high))
        band = Band(
            _x, _y0, _y1, orient=Orientation.HORIZONTAL, color=color, alpha=alpha,
            hatch=hatch, name=f"xband-of-{self.name}", backend=self._backend_name,
        )  # fmt: skip
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LineBand(self, band, name=old_name)

    def with_yband(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        *,
        color: ColorType | _Void = _void,
        alpha: float = 0.3,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _lg.LineBand[Self]:
        from whitecanvas.layers._primitive import Band
        from whitecanvas.layers.group import LineBand

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.color
        data = self.data
        _x, _y0 = self._data_to_backend_data(XYData(data.x, data.y - err))
        _, _y1 = self._data_to_backend_data(XYData(data.x, data.y + err_high))
        band = Band(
            _x, _y0, _y1, orient=Orientation.VERTICAL, color=color, alpha=alpha,
            hatch=hatch, name=f"yband-of-{self.name}", backend=self._backend_name,
        )  # fmt: skip
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LineBand(self, band, name=old_name)

    def with_xfill(
        self,
        bottom: float = 0.0,
        *,
        color: ColorType | _Void = _void,
        alpha: float = 0.3,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _lg.LineBand[Self]:
        from whitecanvas.layers._primitive import Band
        from whitecanvas.layers.group import LineBand

        if color is _void:
            color = self.color
        data = self.data
        x0 = np.full((data.x.size,), bottom)
        band = Band(
            data.y, x0, data.x, orient=Orientation.HORIZONTAL, color=color, alpha=alpha,
            hatch=hatch, name=f"xfill-of-{self.name}", backend=self._backend_name,
        )  # fmt: skip
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LineBand(self, band, name=old_name)

    def with_yfill(
        self,
        bottom: float = 0.0,
        *,
        color: ColorType | _Void = _void,
        alpha: float = 0.3,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _lg.LineBand[Self]:
        from whitecanvas.layers._primitive import Band
        from whitecanvas.layers.group import LineBand

        if color is _void:
            color = self.color
        data = self.data
        y0 = np.full((data.y.size,), bottom)
        band = Band(
            data.x, y0, data.y, orient=Orientation.VERTICAL, color=color, alpha=alpha,
            hatch=hatch, name=f"yfill-of-{self.name}", backend=self._backend_name,
        )  # fmt: skip
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LineBand(self, band, name=old_name)


class Line(_SingleLine):
    _backend_class_name = "MonoLine"
    events: LineLayerEvents
    _events_class = LineLayerEvents

    def __init__(
        self,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        *,
        name: str | None = None,
        color: ColorType = "blue",
        width: float = 1.0,
        alpha: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        xdata, ydata = normalize_xy(xdata, ydata)
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), xdata, ydata)
        self.update(
            color=color, width=width, style=style, alpha=alpha, antialias=antialias
        )
        self._x_hint, self._y_hint = xy_size_hint(xdata, ydata)
        self._backend._plt_connect_pick_event(self.events.clicked.emit)

    def _get_layer_data(self) -> XYData:
        return XYData(*self._backend._plt_get_data())

    def _set_layer_data(self, data: XYData):
        x0, y0 = data
        self._backend._plt_set_data(x0, y0)
        self._x_hint, self._y_hint = xy_size_hint(x0, y0)

    def with_text(
        self,
        strings: list[str],
        *,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        family: str | None = None,
    ) -> _lg.LabeledLine:
        from whitecanvas.layers import Errorbars
        from whitecanvas.layers.group import LabeledLine

        strings = _text_utils.norm_label_text(strings, self.data)
        texts = Texts(
            *self.data, strings, color=color, size=size, rotation=rotation,
            anchor=anchor, family=family, backend=self._backend_name,
        )  # fmt: skip
        old_name = self.name
        self.name = f"line-of-{self.name}"
        return LabeledLine(
            self,
            Errorbars.empty_h(name=f"xerr-of-{self.name}", backend=self._backend_name),
            Errorbars.empty_v(name=f"yerr-of-{self.name}", backend=self._backend_name),
            texts=texts,
            name=old_name,
        )

    def with_hover_template(
        self,
        template: str,
        extra: Any | None = None,
    ) -> Self:
        """Add hover template to the markers."""
        xs, ys = self.data
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, xs.size, extra)
        # set default format keys
        params.setdefault("x", xs)
        params.setdefault("y", ys)
        if "i" not in params:
            params["i"] = np.arange(xs.size)
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(xs.size)
        ]
        self._backend._plt_set_hover_text(texts)
        return self

    @classmethod
    def build_cdf(
        cls,
        data: ArrayLike1D,
        *,
        sorted: bool = False,
        orient: OrientationLike = "vertical",
        name: str | None = None,
        color: ColorType = "blue",
        alpha: float = 1.0,
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        """Construct a line from a cumulative histogram."""
        xdata = as_array_1d(data)
        if not sorted:
            xdata = np.sort(xdata)
        ydata = np.linspace(0, 1, xdata.size)
        if not Orientation.parse(orient).is_vertical:
            xdata, ydata = ydata, xdata
        return Line(
            xdata, ydata, name=name, color=color, alpha=alpha, antialias=antialias,
            width=width, style=style, backend=backend,
        )  # fmt: skip


class LineStep(_SingleLine):
    _backend_class_name = "MonoLine"
    events: LineLayerEvents
    _events_class = LineLayerEvents

    def __init__(
        self,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        *,
        name: str | None = None,
        where: str | StepStyle = StepStyle.MID,
        color: ColorType = "blue",
        width: float = 1,
        alpha: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        xdata, ydata = normalize_xy(xdata, ydata)
        self._where = StepStyle(where)
        super().__init__(name=name)
        xback, yback = self._data_to_backend_data(XYData(xdata, ydata))
        self._backend = self._create_backend(Backend(backend), xback, yback)
        self.update(
            color=color, width=width, style=style, alpha=alpha, antialias=antialias
        )
        self._x_hint, self._y_hint = xy_size_hint(xback, yback)
        self._backend._plt_connect_pick_event(self.events.clicked.emit)

    def _data_to_backend_data(self, data: XYData) -> XYData:
        if data.x.size < 2:
            return data
        if self._where is StepStyle.PRE or self._where is StepStyle.POST_T:
            xdata = np.repeat(data.x, 2)[:-1]
            ydata = np.repeat(data.y, 2)[1:]
        elif self._where is StepStyle.POST or self._where is StepStyle.PRE_T:
            xdata = np.repeat(data.x, 2)[1:]
            ydata = np.repeat(data.y, 2)[:-1]
        elif self._where is StepStyle.MID:
            xrep = np.repeat((data.x[1:] + data.x[:-1]) / 2, 2)
            xdata = np.concatenate([data.x[:1], xrep, data.x[-1:]])
            ydata = np.repeat(data.y, 2)
        elif self._where is StepStyle.MID_T:
            yrep = np.repeat((data.y[1:] + data.y[:-1]) / 2, 2)
            xdata = np.repeat(data.x, 2)
            ydata = np.concatenate([data.y[:1], yrep, data.y[-1:]])
        else:  # pragma: no cover
            raise ValueError(f"Invalid step style: {self._where}")
        return XYData(xdata, ydata)

    def _get_layer_data(self) -> XYData:
        xback, yback = self._backend._plt_get_data()
        if (
            self._where is StepStyle.PRE
            or self._where is StepStyle.POST_T
            or self._where is StepStyle.POST
            or self._where is StepStyle.PRE_T
        ):
            xdata = xback[::2]
            ydata = yback[::2]
        elif self._where is StepStyle.MID:
            xmids = xback[1:-1:2]
            xmid = (xmids[1:] + xmids[:-1]) / 2
            xdata = np.concatenate([xback[:1], xmid, xback[-1:]])
            ydata = yback[1::2]
        elif self._where is StepStyle.MID_T:
            ymids = yback[1:-1:2]
            ymid = (ymids[1:] + ymids[:-1]) / 2
            xdata = xback[::2]
            ydata = np.concatenate([yback[:1], ymid, yback[-1:]])
        else:  # pragma: no cover
            raise ValueError(f"Invalid step style: {self._where}")
        return XYData(xdata, ydata)

    def _norm_layer_data(self, data: Any) -> XYData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 2:
                raise ValueError(f"Expected data to be (N, 2), got {data.shape}")
            xdata, ydata = data[:, 0], data[:, 1]
        else:
            xdata, ydata = data
            if xdata is None:
                xdata = self.data.x
            else:
                xdata = as_array_1d(xdata)
            if ydata is None:
                ydata = self.data.y
            else:
                ydata = as_array_1d(ydata)
        if xdata.size != ydata.size:
            raise ValueError(
                "Expected xdata and ydata to have the same size, "
                f"got {xdata.size} and {ydata.size}"
            )
        return XYData(xdata, ydata)

    def _set_layer_data(self, data: XYData):
        x0, y0 = self._data_to_backend_data(data)
        self._backend._plt_set_data(x0, y0)
        self._x_hint, self._y_hint = xy_size_hint(x0, y0)

    @property
    def where(self) -> StepStyle:
        return self._where

    @where.setter
    def where(self, where: str | StepStyle):
        data = self.data
        self._where = StepStyle(where)
        self._set_layer_data(data)


class MultiLineEvents(LayerEvents):
    color = Signal(np.ndarray)
    width = Signal(float)
    style = Signal(str)
    antialias = Signal(bool)
    clicked = Signal(int)


class MultiLine(HoverableDataBoundLayer[MultiLineProtocol, "list[NDArray[np.number]]"]):
    events: MultiLineEvents
    _events_class = MultiLineEvents

    def __init__(
        self,
        data: list[ArrayLike1D],
        *,
        name: str | None = None,
        color: ColorType = "blue",
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        alpha: float = 1.0,
        backend: Backend | str | None = None,
    ):
        data_normed, xhint, yhint = _norm_data(data)
        super().__init__(name=name)
        self._x_hint, self._y_hint = xhint, yhint
        self._backend = self._create_backend(Backend(backend), data_normed)
        self.update(
            color=color, width=width, style=style, alpha=alpha, antialias=antialias
        )
        self._backend._plt_connect_pick_event(self.events.clicked.emit)

    def _get_layer_data(self) -> list[NDArray[np.number]]:
        """Current data of the layer."""
        return self._backend._plt_get_data()

    def _set_layer_data(self, data: list[NDArray[np.number]]):
        data_norm, x_hint, y_hint = _norm_data(data)
        self._backend._plt_set_data(data)
        self._backend._plt_set_data(data_norm)
        self._x_hint, self._y_hint = x_hint, y_hint

    @property
    def nlines(self) -> int:
        """Number of lines."""
        return len(self._backend._plt_get_data())

    ndata = nlines

    @property
    def color(self) -> NDArray[np.floating]:
        """Color of the line."""
        if self.nlines == 0:
            return np.zeros((0, 4), dtype=np.float32)
        return self._backend._plt_get_edge_color()

    @color.setter
    def color(self, color: ColorType):
        col = as_color_array(color, self.nlines)
        self._backend._plt_set_edge_color(col)
        self.events.color.emit(col)

    @property
    def width(self) -> float:
        """Width of the line."""
        if self.nlines == 0:
            return np.zeros(0, dtype=np.float32)
        return self._backend._plt_get_edge_width()

    @width.setter
    def width(self, width: float | Sequence[float]):
        self._backend._plt_set_edge_width(width)
        self.events.width.emit(width)

    @property
    def style(self) -> EnumArray[LineStyle]:
        """Style of the line."""
        if self.nlines == 0:
            return np.zeros(0, dtype=object)
        return np.array(self._backend._plt_get_edge_style(), dtype=object)

    @style.setter
    def style(self, style: str | LineStyle | Sequence[str | LineStyle]):
        if isinstance(style, (str, LineStyle)):
            s = LineStyle(style)
            self._backend._plt_set_edge_style(s)
            self.events.style.emit(s)
        else:
            styles = [LineStyle(s) for s in style]
            self._backend._plt_set_edge_style(styles)
            self.events.style.emit(styles)

    @property
    def alpha(self) -> NDArray[np.float32]:
        """Alpha channel of each line."""
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        if self.nlines == 0:
            return
        col = self.color.copy()
        col[:, 3] = value
        self.color = col

    @property
    def antialias(self) -> bool:
        """Whether to use antialiasing."""
        return self._backend._plt_get_antialias()

    @antialias.setter
    def antialias(self, antialias: bool) -> None:
        if not isinstance(antialias, bool):
            raise TypeError(f"Expected antialias to be bool, got {type(antialias)}")
        self._backend._plt_set_antialias(antialias)
        self.events.antialias.emit(antialias)

    def update(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        width: float | Sequence[float] | _Void = _void,
        style: str | LineStyle | Sequence[str | LineStyle] | _Void = _void,
        alpha: float | Sequence[float] | _Void = _void,
        antialias: bool | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if width is not _void:
            self.width = width
        if style is not _void:
            self.style = style
        if alpha is not _void:
            self.alpha = alpha
        if antialias is not _void:
            self.antialias = antialias
        return self

    def with_hover_template(
        self,
        template: str,
        extra: Any | None = None,
    ) -> Self:
        """Add hover template to the markers."""
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, self.nlines, extra)
        # set default format keys
        if "i" not in params:
            params["i"] = np.arange(self.nlines)
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(self.nlines)
        ]
        self._backend._plt_set_hover_text(texts)
        return self

    def _as_legend_item(self) -> _legend.LineLegendItem:
        if self.nlines == 0:
            return _legend.EmptyLegendItem()
        return _legend.LineLegendItem(self.color[0], self.width[0], self.style[0])


def _norm_data(data: list[NDArray[np.number]]) -> NDArray[np.number]:
    data_normed: list[NDArray[np.number]] = []
    xmins, xmaxs, ymins, ymaxs = [], [], [], []
    for each in data:
        arr = np.asarray(each)
        if arr.dtype.kind not in "uif":
            raise ValueError(f"Expected data to be numeric, got {arr.dtype}")
        if arr.ndim != 2 or arr.shape[1] != 2:
            raise ValueError(f"Expected data to be (N, 2), got {arr.shape}")
        data_normed.append(arr)
        xmins.append(arr[:, 0].min())
        xmaxs.append(arr[:, 0].max())
        ymins.append(arr[:, 1].min())
        ymaxs.append(arr[:, 1].max())
    if len(data) > 0:
        xhint = min(xmins), max(xmaxs)
        yhint = min(ymins), max(ymaxs)
    else:
        xhint = yhint = None
    return data_normed, xhint, yhint



"""Layer with a dataframe bound to it."""

from __future__ import annotations

from itertools import cycle
from typing import TYPE_CHECKING, Any, Callable, Generic, Iterable, TypeVar, overload

import numpy as np
from cmap import Color, Colormap
from numpy.typing import NDArray

from whitecanvas import layers as _l
from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers import group as _lg
from whitecanvas.layers.tabular import _jitter, _shared
from whitecanvas.layers.tabular import _plans as _p
from whitecanvas.layers.tabular._df_compat import DataFrameWrapper, parse
from whitecanvas.types import (
    ColormapType,
    ColorType,
    Hatch,
    HistBinType,
    KdeBandWidthType,
    LineStyle,
    Orientation,
    OrientationLike,
    Rect,
    Symbol,
)
from whitecanvas.utils.hist import histograms

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas.dataframe._base import CatIterator


_DF = TypeVar("_DF")


class DFLines(_shared.DataFrameLayerWrapper[_lg.LineCollection, _DF], Generic[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        segs: list[np.ndarray],
        categories: list[tuple[Any, ...]],
        color: str | tuple[str, ...] | None = None,
        width: float = 1.0,
        style: str | tuple[str, ...] | None = None,
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        splitby = _shared.join_columns(color, style, source=source)
        self._color_by = _p.ColorPlan.default()
        self._style_by = _p.StylePlan.default()
        self._categories = categories
        self._splitby = splitby
        base = _lg.LineCollection(segs, name=name, backend=backend)
        super().__init__(base, source)
        if color is not None:
            self.update_color(color)
        self.update_width(width)
        if style is not None:
            self.update_style(style)
        self.with_hover_template("\n".join(f"{k}: {{{k}!r}}" for k in self._splitby))

    @classmethod
    def from_table(
        cls,
        df: DataFrameWrapper[_DF],
        x: str | _jitter.JitterBase,
        y: str | _jitter.JitterBase,
        color: str | None = None,
        width: float | None = None,
        style: str | None = None,
        name: str | None = None,
        backend: str | Backend | None = None,
    ) -> DFLines[_DF]:
        splitby = _shared.join_columns(color, style, source=df)
        segs = []
        labels: list[tuple[Any, ...]] = []
        if isinstance(x, _jitter.JitterBase):
            xj = x
        else:
            xj = _jitter.IdentityJitter(x)
        if isinstance(y, _jitter.JitterBase):
            yj = y
        else:
            yj = _jitter.IdentityJitter(y)
        for sl, sub in df.group_by(splitby):
            labels.append(sl)
            segs.append(np.column_stack([xj.map(sub), yj.map(sub)]))
        return DFLines(
            df, segs, labels, name=name, color=color, width=width, style=style,
            backend=backend,
        )  # fmt: skip

    @overload
    def update_color(self, value: ColorType) -> Self: ...

    @overload
    def update_color(
        self,
        by: str | Iterable[str],
        palette: ColormapType | None = None,
    ) -> Self: ...

    def update_color(self, by, /, palette=None):
        """Update the color rule of the layer."""
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot color by a column other than {self._splitby}")
            color_by = _p.ColorPlan.from_palette(cov.columns, palette=palette)
        else:
            color_by = _p.ColorPlan.from_const(Color(cov.value))
        self._base_layer.color = color_by.generate(self._categories, self._splitby)
        self._color_by = color_by
        return self

    def update_width(self, value: float) -> Self:
        """Update width of the lines."""
        self._base_layer.width = value
        return self

    def update_style(self, by: str | Iterable[str], styles=None) -> Self:
        """Update the styling rule of the layer."""
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot style by a column other than {self._splitby}")
            style_by = _p.StylePlan.new(cov.columns, values=styles)
        else:
            style_by = _p.StylePlan.from_const(LineStyle(cov.value))
        self._base_layer.style = style_by.generate(self._categories, self._splitby)
        self._style_by = style_by
        return self

    def update_alpha(self, value: float) -> Self:
        """Update alpha of the lines."""
        for line in self._base_layer:
            line.alpha = value
        return self

    def move(self, dx: float = 0.0, dy: float = 0.0, autoscale: bool = True) -> Self:
        """Add a constant shift to the layer."""
        for layer in self._base_layer:
            old_data = layer.data
            new_data = old_data[0] + dx, old_data[1] + dy
            layer.data = new_data
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Define hover template to the layer."""
        extra = {}
        for i, key in enumerate(self._splitby):
            extra[key] = [row[i] for row in self._categories]
        self.base.with_hover_template(template, extra=extra)
        return self

    def with_markers(
        self,
        *,
        symbol: str | Symbol = Symbol.CIRCLE,
        size: float | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _lg.MainAndOtherLayers[DFLines[_DF], _lg.MarkerCollection]:
        markers = self._base_layer._prep_markers(
            symbol=symbol, size=size, alpha=alpha, hatch=hatch,
        )  # fmt: skip
        return _lg.MainAndOtherLayers([self, markers], name=self.name)

    def _simple_dataframe(self) -> DataFrameWrapper[dict]:
        return _shared.list_to_df(self._categories, self._splitby)

    def _prep_legend_info(
        self,
    ) -> tuple[list[tuple[str, ColorType]], list[tuple[str, LineStyle]]]:
        df = self._simple_dataframe()
        color_entries = self._color_by.to_entries(df)
        style_entries = self._style_by.to_entries(df)
        return color_entries, style_entries

    def _as_legend_item(self) -> _legend.LegendItemCollection | _legend.LineLegendItem:
        colors, styles = self._prep_legend_info()
        ncolor = len(colors)
        nstyle = len(styles)
        widths = self._base_layer.width

        color_default = theme.get_theme().foreground_color
        style_default = LineStyle.SOLID
        if ncolor == 1:
            _, color_default = colors[0]
        if nstyle == 1:
            _, style_default = styles[0]

        if ncolor == 1 and nstyle == 1:
            return _legend.LineLegendItem(color_default, widths[0], style_default)
        items = []
        if ncolor > 1:
            items.append((", ".join(self._color_by.by), _legend.TitleItem()))
            for (label, color), w in zip(colors, widths):
                item = _legend.LineLegendItem(color, w, style_default)
                items.append((label, item))
        if nstyle > 1:
            items.append((", ".join(self._style_by.by), _legend.TitleItem()))
            for (label, style), w in zip(styles, widths):
                item = _legend.LineLegendItem(color_default, w, style)
                items.append((label, item))
        return _legend.LegendItemCollection(items)


class DFHeatmap(_shared.DataFrameLayerWrapper[_lg.LabeledImage, _DF], Generic[_DF]):
    @property
    def cmap(self) -> Colormap:
        """Colormap of the heatmap."""
        return self._base_layer.cmap

    @cmap.setter
    def cmap(self, cmap: ColormapType):
        self._base_layer.cmap = Colormap(cmap)

    @property
    def clim(self) -> tuple[float, float]:
        """Color limits of the heatmap."""
        return self._base_layer.clim

    @clim.setter
    def clim(self, clim: tuple[float, float]):
        self._base_layer.clim = clim

    @classmethod
    def from_array(
        cls,
        src: DataFrameWrapper[_DF],
        arr: np.ndarray,
        name: str | None = None,
        cmap: ColormapType = "gray",
        clim: tuple[float | None, float | None] | None = None,
        backend: Backend | str | None = None,
    ) -> DFHeatmap[_DF]:
        base = _lg.LabeledImage(
            _l.Image(
                arr, name=f"image-of-{name}", cmap=cmap, clim=clim, backend=backend
            ),
            name=name,
        )
        return cls(base, src)

    def with_text(
        self,
        *,
        size: int = 8,
        color_rule: ColorType | Callable[[np.ndarray], ColorType] | None = None,
        fmt: str = "",
        text_invalid: str | None = None,
        mask: NDArray[np.bool_] | None = None,
    ) -> Self:
        """
        Add text layer that displays the pixel values of the heatmap.

        Parameters
        ----------
        size : int, default 8
            Font size of the text.
        color_rule : color-like, callable, optional
            Rule to define the color for each text based on the color-mapped image
            intensity.
        fmt : str, optional
            Format string for the text.
        mask : array_like, optional
            Mask to specify which pixel to add text if specified.
        """
        self._base_layer.with_text(
            size=size, color_rule=color_rule, fmt=fmt, text_invalid=text_invalid,
            mask=mask,
        )  # fmt: skip
        return self

    def with_colorbar(
        self,
        bbox: Rect | None = None,
        orient: OrientationLike = "vertical",
    ) -> Self:
        """Add colorbar to the heatmap."""
        self._base_layer.with_colorbar(bbox=bbox, orient=orient)
        return self


class DFMultiHeatmap(
    _shared.DataFrameLayerWrapper[_lg.LayerCollectionBase[_l.Image], _DF],
    Generic[_DF],
):
    _NO_PADDING_NEEDED = True

    def __init__(
        self,
        base: _lg.LayerCollectionBase[_l.Image],
        source: DataFrameWrapper[_DF],
        color_by: _p.ColorPlan,
        categories: list[tuple],
    ):
        self._color_by = color_by
        self._categories = categories
        super().__init__(base, source)

    @classmethod
    def build_hist(
        cls,
        df: _DF,
        x: str,
        y: str,
        name: str | None = None,
        color: str | list[str] | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        range=None,
        palette: ColormapType = "tab10",
        backend: Backend | str | None = None,
    ) -> Self:
        src, color = cls._norm_df_xy_color(df, [x, y], color)
        # normalize bins
        if isinstance(bins, tuple):
            xbins, ybins = bins
        else:
            xbins = ybins = bins
        if range is None:
            xrange = yrange = None
        else:
            xrange, yrange = range
        _bins = (
            np.histogram_bin_edges(src[x], xbins, xrange),
            np.histogram_bin_edges(src[y], ybins, yrange),
        )

        color_by = _p.ColorPlan.from_palette(color, palette)
        image_layers: list[_l.Image] = []
        categories = []
        color_iter = cycle(color_by.values)
        for sl, sub in src.group_by(color):
            categories.append(sl)
            xdata, ydata = sub[x], sub[y]
            cmap = _gen_cmap_from_color(next(color_iter))
            img = _l.Image.build_hist(
                xdata, ydata, name=f"heatmap-{sl}", cmap=cmap, bins=_bins, density=True,
                backend=backend,
            )  # fmt: skip
            image_layers.append(img)
        base = _lg.LayerCollectionBase(image_layers, name=name)
        return cls(base, src, color_by, categories)

    @classmethod
    def build_kde(
        cls,
        df: _DF,
        x: str,
        y: str,
        name: str | None = None,
        color: str | list[str] | None = None,
        band_width: KdeBandWidthType = "scott",
        palette: ColormapType = "tab10",
        backend: Backend | str | None = None,
    ) -> Self:
        src, color = cls._norm_df_xy_color(df, [x, y], color)
        color_by = _p.ColorPlan.from_palette(color, palette)
        image_layers: list[_l.Image] = []
        categories = []
        color_iter = cycle(color_by.values)
        xrange = src[x].min(), src[x].max()
        yrange = src[y].min(), src[y].max()
        for sl, sub in src.group_by(color):
            categories.append(sl)
            xdata, ydata = sub[x], sub[y]
            cmap = _gen_cmap_from_color(next(color_iter))
            img = _l.Image.build_kde(
                xdata, ydata, name=f"heatmap-{sl}", cmap=cmap, band_width=band_width,
                range=(xrange, yrange), backend=backend,
            )  # fmt: skip
            image_layers.append(img)
        base = _lg.LayerCollectionBase(image_layers, name=name)
        return cls(base, src, color_by, categories)

    @classmethod
    def build_hist_1d(
        cls,
        cat: CatIterator[_DF],
        offsets: str | list[str],
        value: str,
        name: str | None = None,
        color: str | list[str] | None = None,
        dodge: str | list[str] | bool = False,
        bins: HistBinType = "auto",
        range=None,
        palette: ColormapType = "tab10",
        orient: OrientationLike = "vertical",
        backend: Backend | str | None = None,
    ) -> DFMultiHeatmap[_DF]:
        if not isinstance(bins, (int, np.integer, str)):
            raise NotImplementedError("Only equal-width bins are supported.")
        src, color = cls._norm_df_xy_color(cat.df, [value], color)
        ori = Orientation.parse(orient)
        splitby, dodge = _shared.norm_dodge(cat.df, offsets, color, dodge=dodge)
        xs, arrays, categories = cat.prep_arrays(splitby, value, dodge=dodge, width=1.0)
        extent = cat.zoom_factor(dodge, width=1.0)
        hist = histograms(arrays, bins, range)
        ymin, ymax = hist.edges[0], hist.edges[-1]
        image_layers = list[_l.Image]()
        densities = hist.density()
        color_by = _p.ColorPlan.from_palette(color, palette)
        clim = (0, max(dens.max() for dens in densities))
        colors = color_by.generate(categories, splitby)
        for i, dens in enumerate(densities):
            cmap = _gen_cmap_from_color(colors[i])
            x = xs[i]
            if ori.is_vertical:
                img = _l.Image(
                    dens.reshape(-1, 1), cmap=cmap, clim=clim, backend=backend,
                ).fit_to(x - extent / 2, x + extent / 2, ymin, ymax)  # fmt: skip
            else:
                img = _l.Image(
                    dens.reshape(1, -1), cmap=cmap, clim=clim, backend=backend,
                ).fit_to(ymin, ymax, x - extent / 2, x + extent / 2)  # fmt: skip
            image_layers.append(img)
        base = _lg.LayerCollectionBase(image_layers, name=name)
        return cls(base, src, color_by, categories)

    @staticmethod
    def _norm_df_xy_color(df, cols: list[str], color):
        src = parse(df)
        # dtype check
        for x in cols:
            if src[x].dtype.kind not in "fiub":
                raise ValueError(f"Column {x!r} is not numeric.")

        if isinstance(color, str):
            color = (color,)
        elif color is None:
            color = ()
        else:
            color = tuple(color)
        return src, color

    def _as_legend_item(self) -> _legend.LegendItem:
        if len(self._categories) == 1:
            face = _legend.FaceInfo(self._color_by.values[0])
            edge = _legend.EdgeInfo(self._color_by.values[0], width=0)
            return _legend.BarLegendItem(face, edge)
        df = _shared.list_to_df(self._categories, self._color_by.by)
        colors = self._color_by.to_entries(df)
        items = [(", ".join(self._color_by.by), _legend.TitleItem())]
        for label, color in colors:
            face = _legend.FaceInfo(color)
            edge = _legend.EdgeInfo(color, width=0)
            items.append((label, _legend.BarLegendItem(face, edge)))
        return _legend.LegendItemCollection(items)


def _gen_cmap_from_color(next_color: Color):
    next_background = Color([*next_color.rgba[:3], 0.0])
    return [next_background, next_color]


class DFPointPlot2D(_shared.DataFrameLayerWrapper[_lg.LabeledPlot, _DF], Generic[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        x: str,
        y: str,
        *,
        color: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        size: float | None = None,
        capsize: float = 0.15,
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        cols = _shared.join_columns(color, hatch, source=source)
        xdata = []
        ydata = []
        for _, sub in source.group_by(cols):
            xdata.append(sub[x])
            ydata.append(sub[y])
        base = _lg.LabeledPlot.from_arrays_2d(
            xdata, ydata, name=name, capsize=capsize, backend=backend
        )
        if size is not None:
            base.markers.size = size
        super().__init__(base, source)


_L = TypeVar("_L", bound=_lg.LineFillBase)


class DFLineFillBase(
    _shared.DataFrameLayerWrapper[_lg.LayerCollectionBase[_L], _DF],
    Generic[_L, _DF],
):
    _ATTACH_TO_AXIS = True

    def __init__(
        self,
        base: _lg.LayerCollectionBase[_L],
        source: DataFrameWrapper[_DF],
        categories: list[tuple[Any, ...]],
        splitby: tuple[str, ...],
    ):
        self._categories = categories
        self._splitby = splitby
        super().__init__(base, source)
        self._color_by = _p.ColorPlan.default()
        self._width_by = _p.WidthPlan.default()
        self._style_by = _p.StylePlan.default()
        self._hatch_by = _p.HatchPlan.default()

    @overload
    def update_color(self, value: ColorType) -> Self: ...

    @overload
    def update_color(
        self,
        by: str | Iterable[str],
        palette: ColormapType | None = None,
    ) -> Self: ...

    def update_color(self, by, /, palette=None):
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot color by a column other than {self._splitby}")
            color_by = _p.ColorPlan.from_palette(cov.columns, palette=palette)
        else:
            color_by = _p.ColorPlan.from_const(Color(cov.value))
        for i, col in enumerate(color_by.generate(self._categories, self._splitby)):
            self._base_layer[i].color = col
        self._color_by = color_by
        return self

    def update_width(self, value: float) -> Self:
        for hist in self._base_layer:
            hist.line.width = value
        return self

    def update_style(self, by: str | Iterable[str], palette=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot style by a column other than {self._splitby}")
            style_by = _p.StylePlan.new(cov.columns, values=palette)
        else:
            style_by = _p.StylePlan.from_const(LineStyle(cov.value))
        for i, st in enumerate(style_by.generate(self._categories, self._splitby)):
            self._base_layer[i].line.style = st
        self._style_by = style_by
        return self

    def update_hatch(self, by: str | Iterable[str], styles=None) -> Self:
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot hatch by a column other than {self._splitby}")
            hatch_by = _p.HatchPlan.new(cov.columns, values=styles)
        else:
            hatch_by = _p.HatchPlan.from_const(cov.value)
        for i, st in enumerate(hatch_by.generate(self._categories, self._splitby)):
            self._base_layer[i].fill.face.hatch = st
        self._hatch_by = hatch_by
        return self

    def _as_legend_item(self) -> _legend.LegendItemCollection:
        if len(self.base) == 0:
            return _legend.LegendItemCollection([])
        df = _shared.list_to_df(self._categories, self._splitby)
        colors = self._color_by.to_entries(df)
        hatches = self._hatch_by.to_entries(df)
        items = []
        color_default = theme.get_theme().background_color
        fill_alpha = self.base[0].fill_alpha
        width = self.base[0].line.width
        style = self.base[0].line.style
        if self._color_by.is_const():
            color_default = Color(self._color_by.values[0])
        else:
            items.append((", ".join(self._color_by.by), _legend.TitleItem()))
            for label, color in colors:
                fc = Color([*color.rgba[:3], fill_alpha])
                _face = _legend.FaceInfo(fc)
                _edge = _legend.EdgeInfo(color, width, style)
                item = _legend.BarLegendItem(_face, _edge)
                items.append((label, item))
        if self._hatch_by.is_not_const():
            items.append((", ".join(self._hatch_by.by), _legend.TitleItem()))
            for label, hatch in hatches:
                fc = Color([*color_default.rgba[:3], fill_alpha])
                _face = _legend.FaceInfo(fc, hatch)
                _edge = _legend.EdgeInfo(color_default, width, style)
                item = _legend.BarLegendItem(_face, _edge)
                items.append((label, item))
        return _legend.LegendItemCollection(items)


class DFHistograms(DFLineFillBase[_lg.Histogram, _DF], Generic[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        base: _lg.LayerCollectionBase[_lg.Histogram],
        labels: list[tuple[Any, ...]],
        color: str | tuple[str, ...] | None = None,
        width: str | None = None,
        style: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
    ):
        splitby = _shared.join_columns(color, style, source=source)
        super().__init__(base, source, labels, splitby)
        if color is not None:
            self.update_color(color)
        if isinstance(width, str):
            self.update_width(width)
        if style is not None:
            self.update_style(style)
        if hatch is not None:
            self.update_hatch(hatch)

    @property
    def orient(self) -> Orientation:
        if len(self.base) > 0:
            return self.base[0].orient
        return Orientation.VERTICAL

    @classmethod
    def from_table(
        cls,
        df: DataFrameWrapper[_DF],
        value: str,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        kind="count",
        shape="bars",
        color: str | None = None,
        width: float = 1.0,
        style: str | None = None,
        hatch: str | None = None,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        backend: str | Backend | None = None,
    ) -> DFHistograms[_DF]:
        splitby = _shared.join_columns(color, style, source=df)
        ori = Orientation.parse(orient)
        arrays: list[np.ndarray] = []
        labels: list[tuple] = []
        for sl, sub in df.group_by(splitby):
            labels.append(sl)
            arrays.append(sub[value])
        hist = histograms(arrays, bins, limits)

        layers = []
        for arr in arrays:
            each_layer = _lg.Histogram.from_array(
                arr, kind=kind, bins=hist.edges, limits=limits, width=width,
                orient=ori, shape=shape, backend=backend,
            )  # fmt: skip
            layers.append(each_layer)
        base = _lg.LayerCollectionBase(layers, name=name)
        return cls(df, base, labels, color=color, width=width, style=style, hatch=hatch)


class DFKde(DFLineFillBase[_lg.Kde, _DF], Generic[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        base: _lg.LayerCollectionBase[_lg.Kde],
        labels: list[tuple[Any, ...]],
        color: str | tuple[str, ...] | None = None,
        width: str | None = None,
        style: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
    ):
        splitby = _shared.join_columns(color, style, source=source)
        super().__init__(base, source, labels, splitby)
        if color is not None:
            self.update_color(color)
        if isinstance(width, str):
            self.update_width(width)
        if style is not None:
            self.update_style(style)
        if hatch is not None:
            self.update_hatch(hatch)

    @property
    def orient(self) -> Orientation:
        if len(self.base) > 0:
            return self.base[0].orient
        return Orientation.VERTICAL

    @classmethod
    def from_table(
        cls,
        df: DataFrameWrapper[_DF],
        value: str,
        band_width: KdeBandWidthType = "scott",
        color: str | None = None,
        width: float = 1.0,
        style: str | None = None,
        hatch: str | None = None,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        backend: str | Backend | None = None,
    ) -> DFKde[_DF]:
        splitby = _shared.join_columns(color, style, source=df)
        ori = Orientation.parse(orient)
        arrays: list[np.ndarray] = []
        labels: list[tuple] = []
        for sl, sub in df.group_by(splitby):
            labels.append(sl)
            arrays.append(sub[value])
        layers = []
        for arr in arrays:
            each_layer = _lg.Kde.from_array(
                arr, width=width, band_width=band_width, orient=ori, backend=backend,
            )  # fmt: skip
            layers.append(each_layer)
        base = _lg.LayerCollectionBase(layers, name=name)
        return cls(df, base, labels, color=color, width=width, style=style, hatch=hatch)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Sequence

from whitecanvas.backend import Backend
from whitecanvas.canvas import (
    Canvas,
    CanvasGrid,
    CanvasHGrid,
    CanvasVGrid,
    JointGrid,
    SingleCanvas,
)
from whitecanvas.types import ColormapType

if TYPE_CHECKING:
    from typing import Literal

    _0_or_1 = Literal[0, 1]


def new_canvas(
    backend: Backend | str | None = None,
    *,
    size: tuple[int, int] | None = None,
    palette: str | ColormapType | None = None,
) -> SingleCanvas:
    """
    Create a new canvas with a single cell.

    Parameters
    ----------
    backend : Backend or str, optional
        Backend name.
    size : (int, int), optional
        Displaying size of the canvas (in pixels).
    palette : str or ColormapType, optional
        Color palette of the canvas. This color palette will be used to generate colors
        for the plots.
    """
    _grid = CanvasGrid([1], [1], backend=backend)
    _grid.add_canvas(0, 0, palette=palette)
    cvs = SingleCanvas(_grid)
    if size is not None:
        cvs.size = size
    return cvs


def new_canvas_3d(
    backend: Backend | str | None = None,
    *,
    size: tuple[int, int] | None = None,
    palette: str | ColormapType | None = None,
):
    from whitecanvas.canvas.canvas3d._base import SingleCanvas3D

    _grid = CanvasGrid([1], [1], backend=backend)
    _grid.add_canvas_3d(0, 0, palette=palette)
    cvs = SingleCanvas3D(_grid)
    if size is not None:
        cvs.size = size
    return cvs


def new_grid(
    rows: int | Sequence[int] = 1,
    cols: int | Sequence[int] = 1,
    *,
    size: tuple[int, int] | None = None,
    backend: Backend | str | None = None,
) -> CanvasGrid:
    """
    Create a new canvas grid with uniform or non-uniform cell sizes.

    >>> grid = new_grid(2, 3)  # 2x3 grid
    >>> grid = new_grid(2, 3, size=(800, 600))  # 2x3 grid with size 800x600
    >>> grid = new_grid([1, 2], [2, 1])  # 2x2 grid with non-uniform sizes

    If you want to create a 1D grid, use `new_row` or `new_col` instead.

    Parameters
    ----------
    rows : int or sequence of int, default 1
        Number of rows (if an integer is given) or height ratio of the rows (if a
        sequence of intergers is given).
    cols : int or sequence of int, default 1
        Number of columns (if an integer is given) or width ratio of the columns (if a
        sequence of intergers is given).
    size : (int, int), optional
        Displaying size of the grid (in pixels).
    backend : Backend or str, optional
        Backend name, such as "matplotlib:qt".

    Returns
    -------
    CanvasGrid
        Grid of empty canvases.
    """
    heights = _norm_ratio(rows)
    widths = _norm_ratio(cols)
    grid = CanvasGrid(heights, widths, backend=backend)
    if size is not None:
        grid.size = size
    return grid


def new_row(
    cols: int | Sequence[int] = 1,
    *,
    size: tuple[int, int] | None = None,
    backend: Backend | str | None = None,
) -> CanvasHGrid:
    """Create a new horizontal canvas grid with uniform or non-uniform cell sizes."""
    widths = _norm_ratio(cols)
    grid = CanvasHGrid(widths, backend=backend)
    if size is not None:
        grid.size = size
    return grid


def new_col(
    rows: int | Sequence[int] = 1,
    *,
    size: tuple[int, int] | None = None,
    backend: Backend | str | None = None,
) -> CanvasVGrid:
    """Create a new vertical canvas grid with uniform or non-uniform cell sizes."""
    heights = _norm_ratio(rows)
    grid = CanvasVGrid(heights, backend=backend)
    if size is not None:
        grid.size = size
    return grid


def new_jointgrid(
    backend: Backend | str | None = None,
    *,
    loc: tuple[_0_or_1, _0_or_1] = (1, 0),
    size: tuple[int, int] | None = None,
    palette: str | ColormapType | None = None,
) -> JointGrid:
    """
    Create a new joint grid.

    Parameters
    ----------
    backend : Backend or str, optional
        Backend of the canvas.
    loc : (int, int), default (1, 0)
        Location of the main canvas. Each integer must be 0 or 1.
    size : (int, int), optional
        Size of the canvas in pixel.
    palette : colormap type, optional
        Color palette used for the canvases.

    Returns
    -------
    JointGrid
        Joint grid object.
    """
    joint = JointGrid(loc, palette=palette, backend=backend)
    if size is not None:
        joint.size = size
    return joint


def wrap_canvas(obj: Any, palette=None) -> Canvas:
    """
    Wrap a backend object into a whitecanvas Canvas.

    >>> import matplotlib.pyplot as plt
    >>> canvas = wrap_canvas(plt.gca())
    """
    typ = type(obj).__name__

    if _is_in_module(typ, "matplotlib", "Axes"):
        from matplotlib.axes import Axes

        from whitecanvas.backend.matplotlib import Canvas as BackendCanvas

        if not isinstance(obj, Axes):
            raise TypeError(f"Expected matplotlib Axes, got {typ}")
        backend = "matplotlib"

    elif _is_in_module(typ, "plotly", "Figure"):
        from plotly.graph_objs import Figure

        from whitecanvas.backend.plotly import Canvas as BackendCanvas

        if not isinstance(obj, Figure):
            raise TypeError(f"Expected plotly Figure, got {typ}")
        backend = "plotly"
    elif _is_in_module(typ, "bokeh", "figure"):
        from bokeh.plotting import figure

        from whitecanvas.backend.bokeh import Canvas as BackendCanvas

        if not isinstance(obj, figure):
            raise TypeError(f"Expected bokeh figure, got {typ}")
        backend = "bokeh"
    elif _is_in_module(typ, "vispy", "ViewBox"):
        from vispy.scene import ViewBox

        from whitecanvas.backend.vispy import Canvas as BackendCanvas

        if not isinstance(obj, ViewBox):
            raise TypeError(f"Expected vispy ViewBox, got {typ}")
        backend = "vispy"
    elif _is_in_module(typ, "pyqtgraph", "PlotItem"):
        from pyqtgraph import PlotItem

        from whitecanvas.backend.pyqtgraph import Canvas as BackendCanvas

        if not isinstance(obj, PlotItem):
            raise TypeError(f"Expected pyqtgraph PlotItem, got {typ}")
        backend = "pyqtgraph"
    else:
        raise TypeError(f"Cannot convert {typ} to Canvas")
    return Canvas.from_backend(BackendCanvas(obj), palette=palette, backend=backend)


def _is_in_module(typ_str: str, mod_name: str, cls_name: str) -> bool:
    return mod_name in sys.modules and typ_str.split(".")[-1] == cls_name


def _norm_ratio(r: int | Sequence[int]) -> list[int]:
    if hasattr(r, "__int__"):
        out = [1] * int(r)
    else:
        out: list[int] = []
        for x in r:
            if not hasattr(x, "__int__"):
                raise ValueError(f"Invalid value for size ratio: {r!r}.")
            out.append(int(x))
        if len(out) == 0:
            raise ValueError("Size ratio must not be empty.")
    return out



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, TypeVar

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend, _text_utils
from whitecanvas.layers._base import HoverableDataBoundLayer
from whitecanvas.layers._mixin import (
    ConstEdge,
    ConstFace,
    EdgeNamespace,
    FaceEdgeMixinEvents,
    FaceNamespace,
    MultiEdge,
    MultiFace,
    MultiFaceEdgeMixin,
)
from whitecanvas.layers._primitive.text import Texts
from whitecanvas.layers._sizehint import xyy_size_hint
from whitecanvas.protocols import BarProtocol
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    OrientationLike,
    XYData,
    _Void,
)
from whitecanvas.utils.normalize import as_array_1d, parse_texts

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.layers import group as _lg


_void = _Void()
_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)


class BarEvents(FaceEdgeMixinEvents):
    bar_width = Signal(float)
    bottom = Signal(np.ndarray)
    clicked = Signal(int)


class Bars(
    HoverableDataBoundLayer[BarProtocol, XYData],
    MultiFaceEdgeMixin[_Face, _Edge],
):
    """
    Layer that represents vertical/hosizontal bars.

    Attributes
    ----------
    face : `whitecanvas.layers._mixin.FaceNamespace`
        Face properties of the bars.
    edge : `whitecanvas.layers._mixin.EdgeNamespace`
        Edge properties of the bars.
    """

    _ATTACH_TO_AXIS = True
    events: BarEvents
    _events_class = BarEvents

    def __init__(
        self,
        x: ArrayLike1D,
        height: ArrayLike1D,
        bottom: ArrayLike1D | None = None,
        *,
        orient: OrientationLike = "vertical",
        extent: float = 0.8,
        name: str | None = None,
        color: ColorType = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: Backend | str | None = None,
    ):
        MultiFaceEdgeMixin.__init__(self)
        ori = Orientation.parse(orient)
        xxyy, xhint, yhint = _norm_bar_inputs(x, height, bottom, ori, extent)
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), *xxyy)
        self._bar_width = extent
        self._orient = ori
        self.face.update(color=color, alpha=alpha, hatch=hatch)
        self._x_hint, self._y_hint = xhint, yhint
        self._bar_type = "bars"
        self._init_events()
        self._backend._plt_connect_pick_event(self.events.clicked.emit)

    def _get_layer_data(self) -> XYData:
        """Current data of the layer."""
        x0, x1, y0, y1 = self._backend._plt_get_data()
        if self._orient.is_vertical:
            return XYData((x0 + x1) / 2, y1 - y0)
        else:
            return XYData((y0 + y1) / 2, x1 - x0)

    def _norm_layer_data(self, data: Any) -> XYData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 2:
                raise ValueError(
                    "Expected data to be a 2D array of shape (N, 2), "
                    f"got {data.shape}"
                )
            x, height = data[:, 0], data[:, 1]
        else:
            x, height = data
        if x is not None:
            x = as_array_1d(x)
        else:
            x = self.data.x
        if height is not None:
            height = as_array_1d(height)
        else:
            height = self.data.y
        return XYData(x, height)

    def _set_layer_data(self, data: XYData):
        x, height = data
        w = self.bar_width / 2
        if self._orient.is_vertical:
            self._backend._plt_set_data(x - w, x + w, height / 2, -height / 2)
        else:
            self._backend._plt_set_data(-height / 2, height / 2, x - w, x + w)
        self._x_hint, self._y_hint = xyy_size_hint(x, height, height, self.orient)

    def set_data(
        self,
        xdata: ArrayLike1D | None = None,
        ydata: ArrayLike1D | None = None,
        bottom: ArrayLike1D | None = None,
    ):
        emit_data = xdata is not None or ydata is not None
        emit_bottom = bottom is not None
        if xdata is None or ydata is None:
            data = self.data
            if xdata is None:
                xdata = data.x
            if ydata is None:
                ydata = data.y
        if not emit_bottom:
            bottom = self.bottom
        xxyy, xhint, yhint = _norm_bar_inputs(
            xdata, ydata, bottom, self._orient, self._bar_width
        )
        self._backend._plt_set_data(*xxyy)
        self._x_hint, self._y_hint = xhint, yhint
        if emit_data:
            self.events.data.emit(xdata, ydata)
        if emit_bottom:
            self.events.bottom.emit(bottom)

    @property
    def bottom(self) -> NDArray[np.floating]:
        """The bottom values of the bars."""
        x0, _, y0, _ = self._backend._plt_get_data()
        if self._orient.is_vertical:
            return y0
        else:
            return x0

    @bottom.setter
    def bottom(self, bot: ArrayLike1D):
        self.set_data(bottom=bot)

    @property
    def top(self) -> NDArray[np.floating]:
        """The top values of the bars."""
        _, x1, _, y1 = self._backend._plt_get_data()
        if self._orient.is_vertical:
            return y1
        else:
            return x1

    @top.setter
    def top(self, top: ArrayLike1D):
        top = as_array_1d(top)
        self.set_data(ydata=top - self.bottom)

    @property
    def ndata(self) -> int:
        """The number of data points"""
        return self.data.x.size

    @property
    def bar_width(self) -> float:
        """Width of the bars."""
        return self._bar_width

    @bar_width.setter
    def bar_width(self, w: float):
        if w <= 0:
            raise ValueError(f"Expected width > 0, got {w}")
        x0, x1, y0, y1 = self._backend._plt_get_data()
        if self._orient.is_vertical:
            dx = (w - self._bar_width) / 2
            x0 = x0 - dx
            x1 = x1 + dx
        else:
            dy = (w - self._bar_width) / 2
            y0 = y0 - dy
            y1 = y1 + dy
        self._backend._plt_set_data(x0, x1, y0, y1)
        self._bar_width = w
        self.events.bar_width.emit(w)

    @property
    def orient(self) -> Orientation:
        """Orientation of the bars."""
        return self._orient

    def with_err(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = True,
        capsize: float = 0,
    ) -> _lg.LabeledBars:
        if self.orient.is_vertical:
            return self.with_yerr(
                err, err_high, color=color, width=width,
                style=style, antialias=antialias, capsize=capsize
            )  # fmt: skip
        else:
            return self.with_xerr(
                err, err_high, color=color, width=width,
                style=style, antialias=antialias, capsize=capsize
            )  # fmt: skip

    def with_xerr(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = True,
        capsize: float = 0,
    ) -> _lg.LabeledBars:
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group import LabeledBars

        old_name = self.name
        xerr = self._create_errorbars(
            err, err_high, color=color, width=width, style=style,
            antialias=antialias, capsize=capsize, orient=Orientation.HORIZONTAL,
        )  # fmt: skip
        yerr = Errorbars.empty_h(f"xerr-of-{old_name}", backend=self._backend_name)
        self.name = old_name
        return LabeledBars(self, xerr, yerr, name=old_name)

    def with_yerr(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool = True,
        capsize: float = 0,
    ) -> _lg.LabeledBars:
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group import LabeledBars

        old_name = self.name
        yerr = self._create_errorbars(
            err, err_high, color=color, width=width, style=style,
            antialias=antialias, capsize=capsize, orient="vertical",
        )  # fmt: skip
        xerr = Errorbars.empty_v(name=f"yerr-of-{old_name}", backend=self._backend_name)
        self.name = old_name
        return LabeledBars(self, xerr, yerr, name=old_name)

    def with_text(
        self,
        strings: list[str],
        *,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        fontfamily: str | None = None,
    ) -> _lg.LabeledBars:
        from whitecanvas.layers import Errorbars
        from whitecanvas.layers.group import LabeledBars

        strings = _text_utils.norm_label_text(strings, self.data)
        texts = Texts(
            *self.data, strings, color=color, size=size, rotation=rotation,
            anchor=anchor, family=fontfamily, backend=self._backend_name,
        )  # fmt: skip
        return LabeledBars(
            self,
            Errorbars.empty_h(name=f"xerr-of-{self.name}", backend=self._backend_name),
            Errorbars.empty_v(name=f"xerr-of-{self.name}", backend=self._backend_name),
            texts=texts,
            name=self.name,
        )

    def _create_errorbars(
        self,
        err: ArrayLike1D,
        err_high: ArrayLike1D | None = None,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool = True,
        capsize: float = 0,
        orient: OrientationLike = "vertical",
    ):
        from whitecanvas.layers._primitive import Errorbars

        orient = Orientation.parse(orient)
        if err_high is None:
            err_high = err
        if color is _void:
            color = self.edge.color
        if width is _void:
            width = self.edge.width
        if style is _void:
            style = self.edge.style
        x, y = self.data
        y = y + self.bottom
        if orient is not self.orient:
            x, y = y, x
        if orient.is_vertical:
            name = f"yerr-of-{self.name}"
        else:
            name = f"xerr-of-{self.name}"
        return Errorbars(
            x, y - err, y + err_high, color=color, width=width, style=style,
            antialias=antialias, capsize=capsize, orient=orient, name=name,
            backend=self._backend_name
        )  # fmt: skip

    def with_face(
        self,
        *,
        color: ColorType | _Void = _void,
        hatch: Hatch | str = Hatch.SOLID,
        alpha: float = 1,
    ) -> Bars[ConstFace, _Edge]:
        return super().with_face(color=color, hatch=hatch, alpha=alpha)

    def with_face_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | _Void = _void,
        hatch: str | Hatch | Sequence[str | Hatch] | _Void = _void,
        alpha: float = 1,
    ) -> Bars[MultiFace, _Edge]:
        return super().with_face_multi(color=color, hatch=hatch, alpha=alpha)

    def with_edge(
        self,
        *,
        color: ColorType | None = None,
        width: float = 1,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Bars[_Face, ConstEdge]:
        return super().with_edge(color=color, width=width, style=style, alpha=alpha)

    def with_edge_multi(
        self,
        *,
        color: ColorType | Sequence[ColorType] | None = None,
        width: float | Sequence[float] = 1,
        style: str | LineStyle | list[str | LineStyle] = LineStyle.SOLID,
        alpha: float = 1,
    ) -> Bars[_Face, MultiEdge]:
        return super().with_edge_multi(
            color=color, width=width, style=style, alpha=alpha
        )

    def as_edge_only(
        self,
        *,
        width: float = 3.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Self:
        """
        Convert the bars to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        color = self.face.color
        if color.ndim == 0:
            pass
        elif color.ndim == 1:
            self.with_edge(color=color, width=width, style=style)
        elif color.ndim == 2:
            self.with_edge_multi(color=color, width=width, style=style)
        else:
            raise RuntimeError("Unreachable error.")
        self.face.update(alpha=0.0)
        return self

    def with_hover_template(self, template: str, extra: Any | None = None) -> Bars:
        xs, ys = self.data
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, xs.size, extra)
        # set default format keys
        params.setdefault("x", xs)
        params.setdefault("y", ys)
        if "i" not in params:
            params["i"] = np.arange(xs.size)
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(xs.size)
        ]
        self._backend._plt_set_hover_text(texts)
        return self

    def _as_legend_item(self) -> _legend.BarLegendItem:
        return _legend.BarLegendItem(
            self.face._as_legend_info(), self.edge._as_legend_info()
        )


def _norm_bar_inputs(t0, height, bot, orient: Orientation, bar_width: float):
    t0 = as_array_1d(t0)
    height = as_array_1d(height)
    if bot is None:
        bot = np.zeros_like(t0)
    bot = as_array_1d(bot)
    if not (t0.size == height.size == bot.size):
        raise ValueError(
            "Expected all arrays to have the same size, "
            f"got {t0.size}, {height.size}, {bot.size}"
        )
    y0 = height + bot
    x_hint, y_hint = xyy_size_hint(t0, y0, bot, orient, bar_width / 2)

    if orient.is_vertical:
        dx = bar_width / 2
        x0, x1 = t0 - dx, t0 + dx
        y0, y1 = bot, y0
    else:
        dy = bar_width / 2
        x0, x1 = bot, y0
        y0, y1 = t0 - dy, t0 + dy
    return (x0, x1, y0, y1), x_hint, y_hint



from __future__ import annotations

import numpy as np
from vispy.scene import visuals

from whitecanvas.backend.vispy.line import MonoLine as MonoLine2D


class MonoLine3D(MonoLine2D):
    def __init__(self, xdata, ydata, zdata):
        data = np.stack([xdata, ydata, zdata], axis=1)
        visuals.Line.__init__(self, data, antialias=True)
        self.unfreeze()

    ##### BaseProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        pos = self.pos
        return pos[:, 0], pos[:, 1], pos[:, 2]

    def _plt_set_data(self, xdata, ydata, zdata):
        self.set_data(np.stack([xdata, ydata, zdata], axis=1))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, Iterable, TypeVar, overload

import numpy as np
from cmap import Colormap
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend, _mixin, _text_utils
from whitecanvas.layers._base import PrimitiveLayer
from whitecanvas.layers._primitive import Bars, Errorbars, Image, Line, Markers, Texts
from whitecanvas.layers._primitive.line import _SingleLine
from whitecanvas.layers.group._cat_utils import check_array_input
from whitecanvas.layers.group._collections import LayerContainer, RichContainerEvents
from whitecanvas.layers.group._offsets import NoOffset, TextOffset
from whitecanvas.layers.group.colorbar import Colorbar
from whitecanvas.layers.group.line_markers import Plot
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    OrientationLike,
    Origin,
    Rect,
    XYData,
)
from whitecanvas.utils.normalize import as_any_1d_array, as_color_array

if TYPE_CHECKING:
    from typing_extensions import Self

_NFace = TypeVar("_NFace", bound="_mixin.FaceNamespace")
_NEdge = TypeVar("_NEdge", bound="_mixin.EdgeNamespace")
_Size = TypeVar("_Size")
_L = TypeVar("_L", bound=_SingleLine)


class _LabeledLayerBase(LayerContainer):
    def __init__(
        self,
        layer: PrimitiveLayer,
        xerr: Errorbars,
        yerr: Errorbars,
        texts: Texts | None = None,
        name: str | None = None,
        offset: TextOffset | None = None,
    ):
        if offset is None:
            offset = NoOffset()
        if texts is None:
            px, py = self._get_data_xy(layer)
            texts = Texts(
                px,
                py,
                [""] * px.size,
                backend=layer._backend_name,
            )
        super().__init__([layer, xerr, yerr, texts], name=name)
        self._text_offset = offset

    def _get_data_xy(
        self, layer: PrimitiveLayer | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        if layer is None:
            layer = self._children[0]
        return layer.data

    def _default_ordering(self, n: int) -> list[int]:
        assert n == 4
        return [2, 0, 1, 3]

    @property
    def xerr(self) -> Errorbars:
        """The errorbars layer for x."""
        return self._children[1]

    @property
    def yerr(self) -> Errorbars:
        """The errorbars layer for y."""
        return self._children[2]

    @property
    def texts(self) -> Texts:
        """The text group layer."""
        return self._children[3]

    @property
    def data(self) -> XYData:
        """The internal (x, y) data of this layer."""
        return self._children[0].data

    def set_data(self, xdata=None, ydata=None):
        """Set the (x, y) data of this layer."""
        px, py = self._get_data_xy()
        if xdata is None:
            dx = 0
        else:
            dx = xdata - px
        if ydata is None:
            dy = 0
        else:
            dy = ydata - py
        self._children[0].set_data(xdata, ydata)
        if self.xerr.ndata > 0:
            y, x0, x1 = self.xerr.data
            self.xerr.set_data(y + dy, x0 + dx, x1 + dx)
        if self.yerr.ndata > 0:
            x, y0, y1 = self.yerr.data
            self.yerr.set_data(x + dx, y0 + dy, y1 + dy)
        if self.texts.ndata > 0:
            dx, dy = self._text_offset._asarray()
            self.texts.set_pos(px + dx, py + dy)

    @property
    def text_offset(self) -> TextOffset:
        """Return the text offset."""
        return self._text_offset

    def with_text_offset(self, dx: Any, dy: Any):
        """Add offset to text positions."""
        _offset = self._text_offset._add(dx, dy)
        if self.texts.ndata > 0:
            px, py = self._get_data_xy()
            xoff, yoff = _offset._asarray()
            self.texts.set_pos(px + xoff, py + yoff)
        self._text_offset = _offset

    def with_xerr(
        self,
        len_lower: float,
        len_higher: float | None = None,
        *,
        color: ColorType = "black",
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        antialias: bool = True,
        capsize: float = 0,
    ) -> Self:
        """
        Set the x error bar data.

        Parameters
        ----------
        len_lower : float
            Length of lower error.
        len_higher : float, optional
            Length of higher error. If not given, set to the same as `len_lower`.
        """
        if len_higher is None:
            len_higher = len_lower
        x, y = self.data
        self.xerr.set_data(y, x - len_lower, x + len_higher)
        self.xerr.update(
            color=color, width=width, style=style, antialias=antialias,
            capsize=capsize,
        )  # fmt: skip
        if canvas := self._canvas_ref():
            canvas._autoscale_for_layer(self.xerr, maybe_empty=False)
        return self

    def with_yerr(
        self,
        len_lower: float,
        len_higher: float | None = None,
        *,
        color: ColorType = "black",
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
        antialias: bool = True,
        capsize: float = 0,
    ) -> Self:
        """
        Set the y error bar data.

        Parameters
        ----------
        len_lower : float
            Length of lower error.
        len_higher : float, optional
            Length of higher error. If not given, set to the same as `len_lower`.
        """
        if len_higher is None:
            len_higher = len_lower
        x, y = self.data
        self.yerr.set_data(x, y - len_lower, y + len_higher)
        self.yerr.update(
            color=color, width=width, style=style, antialias=antialias,
            capsize=capsize
        )  # fmt: skip
        if canvas := self._canvas_ref():
            canvas._autoscale_for_layer(self.yerr, maybe_empty=False)
        return self

    def with_text(
        self,
        strings: str | list[str],
        *,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        family: str | None = None,
        offset: tuple[Any, Any] | None = None,
    ) -> Self:
        """
        Add texts to the layer.

        Parameters
        ----------
        strings : str or list of str
            The text strings. If a single string is given, it will be used for all
            the data points. You can also use format strings with "{x}", "{y}", and
            "{i}" to format the text with the data point values and index.
        color : ColorType, default "black"
            Text color.
        size : float, default 12
            Font point size of the text.
        rotation : float, default 0.0
            Rotation of the text in degrees.
        anchor : str or Alignment, default Alignment.BOTTOM_LEFT
            Text anchoring position.
        family : str, optional
            The font family of the text.
        offset : tuple, default None
            The offset of the text from the data point.
        """
        strings = _text_utils.norm_label_text(strings, self.data)
        if offset is None:
            _offset = self._text_offset
        else:
            _offset = NoOffset()._add(*offset)

        xdata, ydata = self.data
        dx, dy = _offset._asarray()
        self.texts.string = strings
        self.texts.set_pos(xdata + dx, ydata + dy)
        self.texts.update(
            color=color,
            size=size,
            rotation=rotation,
            anchor=anchor,
            family=family,
        )
        return self


class LabeledLine(_LabeledLayerBase, Generic[_L]):
    @property
    def line(self) -> _L:
        """The line layer."""
        return self._children[0]

    def _as_legend_item(self) -> _legend.LineErrorLegendItem:
        line = self.line._as_legend_item()
        if self.xerr.nlines == 0:
            xerr = None
        else:
            xerr = self.xerr._as_legend_item()
        if self.yerr.nlines == 0:
            yerr = None
        else:
            yerr = self.yerr._as_legend_item()
        return _legend.LineErrorLegendItem(line, xerr, yerr)


class LabeledMarkers(_LabeledLayerBase, Generic[_NFace, _NEdge, _Size]):
    @property
    def markers(self) -> Markers[_NFace, _NEdge, _Size]:
        return self._children[0]

    def _as_legend_item(self) -> _legend.MarkerErrorLegendItem:
        markers = self.markers._as_legend_item()
        if self.xerr.nlines == 0:
            xerr = None
        else:
            xerr = self.xerr._as_legend_item()
        if self.yerr.nlines == 0:
            yerr = None
        else:
            yerr = self.yerr._as_legend_item()
        return _legend.MarkerErrorLegendItem(markers, xerr, yerr)


def _init_mean_sd(x, data, color):
    x, data = check_array_input(x, data)
    color = as_color_array(color, len(x))

    est_data = []
    err_data = []

    for sub_data in data:
        _mean = np.mean(sub_data)
        _sd = np.std(sub_data, ddof=1)
        est_data.append(_mean)
        err_data.append(_sd)

    est_data = np.array(est_data)
    err_data = np.array(err_data)
    return x, est_data, err_data


def _init_error_bars(
    x,
    est,
    err,
    orient,
    capsize,
    backend,
) -> tuple[Errorbars, Errorbars]:
    ori = Orientation.parse(orient)
    errorbar = Errorbars(
        x, est - err, est + err, orient=ori, backend=backend, capsize=capsize,
    )  # fmt: skip
    if ori.is_vertical:
        xerr = Errorbars.empty_h(backend=backend)
        yerr = errorbar
    else:
        xerr = errorbar
        yerr = Errorbars.empty_v(backend=backend)
    return xerr, yerr


class LabeledBars(
    _LabeledLayerBase,
    _mixin.AbstractFaceEdgeMixin["PlotFace", "PlotEdge"],
    Generic[_NFace, _NEdge],
):
    _ATTACH_TO_AXIS = True
    events: RichContainerEvents
    _events_class = RichContainerEvents

    def __init__(
        self,
        layer: Bars[_NFace, _NEdge],
        xerr: Errorbars,
        yerr: Errorbars,
        texts: Texts | None = None,
        name: str | None = None,
        offset: TextOffset | None = None,
    ):
        _LabeledLayerBase.__init__(self, layer, xerr, yerr, texts, name, offset)
        _mixin.AbstractFaceEdgeMixin.__init__(self, PlotFace(self), PlotEdge(self))
        self._init_events()

    @property
    def bars(self) -> Bars[_NFace, _NEdge]:
        """The bars layer."""
        return self._children[0]

    @property
    def orient(self) -> Orientation:
        """The orientation of the bars."""
        return self.bars.orient

    def _main_object_layer(self):
        return self.bars

    def _get_data_xy(self, layer: Bars | None = None) -> tuple[np.ndarray, np.ndarray]:
        if layer is None:
            layer = self.bars
        return layer.data.x, layer.top

    def _default_ordering(self, n: int) -> list[int]:
        assert n == 4
        return [0, 1, 2, 3]

    @classmethod
    def from_arrays(
        cls,
        x: list[float],
        data: list[ArrayLike1D],
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        capsize: float = 0.15,
        color: ColorType | list[ColorType] = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        extent: float = 0.8,
        backend: str | Backend | None = None,
    ) -> LabeledBars[_mixin.MultiFace, _mixin.MonoEdge]:
        x, height, err_data = _init_mean_sd(x, data, color)
        bars = Bars(x, height, extent=extent, backend=backend).with_face_multi(
            color=color, hatch=hatch, alpha=alpha
        )
        xerr, yerr = _init_error_bars(x, height, err_data, orient, capsize, backend)
        return cls(bars, xerr=xerr, yerr=yerr, name=name)

    def _as_legend_item(self):
        return self.bars._as_legend_item()


class LabeledPlot(
    _LabeledLayerBase,
    _mixin.AbstractFaceEdgeMixin["PlotFace", "PlotEdge"],
    Generic[_NFace, _NEdge, _Size, _L],
):
    evens: RichContainerEvents
    _events_class = RichContainerEvents

    def __init__(
        self,
        layer: Plot[_L],
        xerr: Errorbars,
        yerr: Errorbars,
        texts: Texts | None = None,
        name: str | None = None,
        offset: TextOffset | None = None,
    ):
        _LabeledLayerBase.__init__(self, layer, xerr, yerr, texts, name, offset)
        _mixin.AbstractFaceEdgeMixin.__init__(self, PlotFace(self), PlotEdge(self))
        self._init_events()

    @property
    def plot(self) -> Plot[_L]:
        """The plot (line + markers) layer."""
        return self._children[0]

    @property
    def line(self) -> _L:
        """The line layer."""
        return self.plot.line

    @property
    def markers(self) -> Markers[_NFace, _NEdge, _Size]:
        """The markers layer."""
        return self.plot.markers

    def _main_object_layer(self):
        """The main layer with face that will be used in PlotFace/PlotEdge."""
        return self.markers

    @classmethod
    def from_arrays(
        cls,
        x: list[float],
        data: list[ArrayLike1D],
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        capsize: float = 0.15,
        color: ColorType | list[ColorType] = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ) -> LabeledPlot[_mixin.MultiFace, _mixin.MultiEdge, float, Line]:
        x, y, err_data = _init_mean_sd(x, data, color)
        xerr, yerr = _init_error_bars(x, y, err_data, orient, capsize, backend)
        if not orient.is_vertical:
            x, y = y, x
        markers = Markers(
            x, y, backend=backend,
        ).with_face_multi(
            color=color, hatch=hatch, alpha=alpha,
        )  # fmt: skip
        lines = Line(x, y, backend=backend)
        plot = Plot(lines, markers)
        lines.visible = False
        return cls(plot, xerr=xerr, yerr=yerr, name=name)

    @classmethod
    def from_arrays_2d(
        cls,
        xdata: list[ArrayLike1D],
        ydata: list[ArrayLike1D],
        *,
        name: str | None = None,
        capsize: float = 0.15,
        color: ColorType | list[ColorType] = "blue",
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ) -> LabeledPlot[_mixin.MultiFace, _mixin.MultiEdge, float, Line]:
        def _estimate(arrs: list[NDArray[np.number]]):
            _mean = []
            _sd = []
            for arr in arrs:
                _mean.append(np.mean(arr))
                _sd.append(np.std(arr, ddof=1))
            return np.array(_mean), np.array(_sd)

        xmean, xsd = _estimate(xdata)
        ymean, ysd = _estimate(ydata)
        markers = Markers(
            xmean,
            ymean,
            backend=backend,
        ).with_face_multi(
            color=color,
            hatch=hatch,
            alpha=alpha,
        )
        lines = Line(xmean, ymean, backend=backend)
        plot = Plot(lines, markers)
        lines.visible = False
        xerr = Errorbars(
            ymean,
            xmean - xsd,
            xmean + xsd,
            orient=Orientation.HORIZONTAL,
            capsize=capsize,
            backend=backend,
        )
        yerr = Errorbars(
            xmean,
            ymean - ysd,
            ymean + ysd,
            orient=Orientation.VERTICAL,
            capsize=capsize,
            backend=backend,
        )
        return cls(plot, xerr=xerr, yerr=yerr, name=name)

    def _as_legend_item(self) -> _legend.LegendItem:
        if self.xerr.nlines == 0:
            xerr = None
        else:
            xerr = self.xerr._as_legend_item()
        if self.yerr.nlines == 0:
            yerr = None
        else:
            yerr = self.yerr._as_legend_item()
        return _legend.PlotErrorLegendItem(self.plot._as_legend_item(), xerr, yerr)


class PlotFace(_mixin.MultiPropertyFaceBase):
    _layer: LabeledPlot[_mixin.MultiFace, _mixin.MultiEdge, float, _L]

    @property
    def color(self) -> NDArray[np.floating]:
        """Face color of the bar."""
        return self._layer._main_object_layer().face.color

    @color.setter
    def color(self, color):
        ndata = self._layer._main_object_layer().ndata
        col = as_color_array(color, ndata)
        self._layer._main_object_layer().with_face_multi(color=col)
        self.events.color.emit(col)

    @property
    def hatch(self) -> _mixin.EnumArray[Hatch]:
        """Face fill hatch."""
        return self._layer._main_object_layer().face.hatch

    @hatch.setter
    def hatch(self, hatch: str | Hatch | Iterable[str | Hatch]):
        ndata = self._layer._main_object_layer().ndata
        hatches = as_any_1d_array(hatch, ndata, dtype=object)
        self._layer._main_object_layer().with_face_multi(hatch=hatches)
        self.events.hatch.emit(hatches)

    @property
    def alpha(self) -> float:
        """Alpha value of the face."""
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color


class PlotEdge(_mixin.MultiPropertyEdgeBase):
    _layer: LabeledPlot[_NFace, _NEdge, float, _L] | LabeledBars[_NFace, _NEdge]

    @property
    def color(self) -> NDArray[np.floating]:
        """Edge color of the plot."""
        return self._layer._main_object_layer().edge.color

    @color.setter
    def color(self, color: ColorType):
        self._layer._main_object_layer().with_edge_multi(color=color)
        if self._layer.xerr.ndata > 0:
            self._layer.xerr.color = color
        if self._layer.yerr.ndata > 0:
            self._layer.yerr.color = color
        self.events.color.emit(color)

    @property
    def width(self) -> NDArray[np.float32]:
        """Edge widths."""
        return self._layer._main_object_layer().edge.width

    @width.setter
    def width(self, width: float):
        self._layer._main_object_layer().edge.width = width
        if self._layer.xerr.ndata > 0:
            self._layer.xerr.width = width
        if self._layer.yerr.ndata > 0:
            self._layer.yerr.width = width
        self.events.width.emit(width)

    @property
    def style(self) -> _mixin.EnumArray[LineStyle]:
        """Edge styles."""
        return self._layer._main_object_layer().edge.style

    @style.setter
    def style(self, style: str | LineStyle):
        style = LineStyle(style)
        self._layer._main_object_layer().edge.style = style
        if self._layer.xerr.ndata > 0:
            self._layer.xerr.style = style
        if self._layer.yerr.ndata > 0:
            self._layer.yerr.style = style
        self.events.style.emit(style)

    @property
    def alpha(self) -> float:
        return self.color[:, 3]

    @alpha.setter
    def alpha(self, value):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color


class LabeledImage(LayerContainer):
    """
    Layer group for an image with texts and colorbar.
    """

    def __init__(
        self,
        layer: Image,
        texts: Texts | None = None,
        colorbar: Colorbar | None = None,
        name: str | None = None,
    ):
        if texts is None:
            texts = Texts(
                [], [], [], name="texts", backend=layer._backend_name
            ).with_font_multi()
        if colorbar is None:
            colorbar = Colorbar(layer.cmap)
            colorbar.visible = False
        layer.events.cmap.connect_setattr(colorbar, "cmap", maxargs=1)
        super().__init__([layer, texts, colorbar], name=name)

    @property
    def image(self) -> Image:
        """The image layer."""
        return self._children[0]

    @property
    def texts(self) -> Texts:
        """The text layer for the overlay texts."""
        return self._children[1]

    @property
    def colorbar(self) -> Colorbar:
        """The colorbar layer."""
        return self._children[2]

    @property
    def cmap(self) -> Colormap:
        """Current colormap."""
        return self.image.cmap

    @cmap.setter
    def cmap(self, cmap: ColormapType):
        self.image.cmap = cmap

    @property
    def clim(self) -> tuple[float, float]:
        """Current contrast limits."""
        return self.image.clim

    @clim.setter
    def clim(self, clim: tuple[float | None, float | None] | None):
        self.image.clim = clim

    @property
    def shape(self) -> tuple[int, int]:
        """The visual shape of the image (shape without the color axis)."""
        return self.image.shape

    @property
    def data(self) -> NDArray[np.number]:
        """The colored image data (N, M, 4)."""
        return self.image.data

    @data.setter
    def data(self, data: NDArray[np.number]):
        self.image.data = data

    @property
    def data_mapped(self) -> NDArray[np.number]:
        """The colored image data (N, M, 4) mapped by the colormap."""
        return self.image.data_mapped

    @property
    def shift(self) -> tuple[float, float]:
        """Current shift from the origin."""
        return self.image.shift

    @shift.setter
    def shift(self, shift: tuple[float, float]):
        self.image.shift = shift

    @property
    def shift_raw(self) -> tuple[float, float]:
        """Current shift from the origin as a raw data."""
        return self.image.shift

    @property
    def scale(self) -> tuple[float, float]:
        """Current scale."""
        return self.image.scale

    @scale.setter
    def scale(self, scale: float | tuple[float, float]):
        self.image.scale = scale

    @property
    def origin(self) -> Origin:
        """Current origin of the image."""
        return self.image.origin

    @origin.setter
    def origin(self, origin: Origin | str):
        self.image.origin = origin

    @overload
    def fit_to(self, bbox: Rect | tuple[float, float, float, float], /) -> Image: ...

    @overload
    def fit_to(
        self, left: float, right: float, bottom: float, top: float, /
    ) -> Image: ...

    def fit_to(self, *args) -> Image:
        """Fit the image to the given bounding box."""
        self.image.fit_to(*args)

    def with_text(
        self,
        *,
        size: int = 8,
        color_rule: ColorType | Callable[[np.ndarray], ColorType] | None = None,
        fmt: str = "",
        text_invalid: str | None = None,
        mask: NDArray[np.bool_] | None = None,
    ) -> LabeledImage:
        """
        Add text annotation to each pixel of the image.

        Parameters
        ----------
        size : int, default 8
            Font size of the text.
        color_rule : color-like, callable, optional
            Rule to define the color for each text based on the color-mapped image
            intensity.
        fmt : str, optional
            Format string for the text.
        mask : array-like, optional
            Mask to specify the valid pixels for the text annotation.
        """
        texts = self.image._make_text_layer(
            size=size, color_rule=color_rule, fmt=fmt,
            text_invalid=text_invalid, mask=mask,
        )  # fmt: skip
        self.texts.data = texts.data
        self.texts.update(color=texts.color, size=texts.size, anchor=texts.anchor)
        return self

    def with_colorbar(
        self,
        bbox: Rect | None = None,
        *,
        orient: OrientationLike = "vertical",
    ) -> LabeledImage:
        """
        Add a colorbar to the image.

        Parameters
        ----------
        bbox : four float, optional
            Bounding box of the colorbar. If `None`, the colorbar will be placed at the
            bottom-right corner of the image.
        orient : str or Orientation, default "vertical"
            Orientation of the colorbar.
        """
        cbar = self.image._make_colorbar(bbox, orient=orient)
        self.colorbar.visible = True
        self.colorbar.fit_to(cbar.lut.bbox)
        return self



import sys
from pathlib import Path
import tempfile
import numpy as np
from numpy.testing import assert_allclose
import matplotlib.pyplot as plt

import pytest
import whitecanvas as wc
from whitecanvas import new_canvas, wrap_canvas, new_canvas_3d
from whitecanvas.types import Point, MouseEvent

from ._utils import assert_color_equal, filter_warning


def test_namespaces(backend: str):
    canvas = new_canvas(backend=backend)
    canvas.aspect_ratio
    canvas.aspect_ratio = 1
    assert canvas.aspect_ratio == 1

    canvas.background_color

    canvas.title.text = "Title-0"
    assert canvas.title.text == "Title-0"
    canvas.title.color = "red"
    assert_color_equal(canvas.title.color, "red")
    canvas.title.size = 20
    assert canvas.title.size == 20
    canvas.title.family = "Arial"
    assert canvas.title.family == "Arial"
    canvas.title.visible = False
    assert not canvas.title.visible
    canvas.title.visible = True
    assert canvas.title.visible

    canvas.x.label.text = "X-Label-0"
    assert canvas.x.label.text == "X-Label-0"
    canvas.x.label.color = "red"
    assert_color_equal(canvas.x.label.color, "red")
    canvas.x.label.size = 20
    assert canvas.x.label.size == 20
    canvas.x.label.family = "Arial"
    assert canvas.x.label.family == "Arial"
    canvas.x.label.visible = False
    assert not canvas.x.label.visible
    canvas.x.label.visible = True
    assert canvas.x.label.visible

    canvas.y.label.text = "Y-Label-0"
    assert canvas.y.label.text == "Y-Label-0"
    canvas.y.label.color = "red"
    assert_color_equal(canvas.y.label.color, "red")
    canvas.y.label.size = 20
    assert canvas.y.label.size == 20
    canvas.y.label.family = "Arial"
    assert canvas.y.label.family == "Arial"

    # test errors
    with pytest.raises(ValueError):
        canvas.x.lim = (1, 0)
    with pytest.raises(TypeError):
        canvas.x.label = 0

    repr(canvas.x)
    repr(canvas.title)
    repr(canvas.x.ticks)

    if backend != "pyqtgraph":  # not implemented in pyqtgraph
        canvas.x.ticks.rotation = 45
        assert canvas.x.ticks.rotation == pytest.approx(45)

    canvas.x.ticks.visible = False
    assert not canvas.x.ticks.visible
    canvas.x.ticks.visible = True
    assert canvas.x.ticks.visible
    canvas.x.ticks.set_labels([0, 1, 2], ["a", "b", "c"])

    # get tick positions and labels are still hard to implement
    if backend in ("mock", "pyqtgraph", "plotly"):
        assert_allclose(canvas.x.ticks.pos, [0, 1, 2])
        assert canvas.x.ticks.labels == ["a", "b", "c"]
    canvas.x.ticks.reset_labels()

    canvas.x.set_gridlines()

def test_namespace_pointing_at_different_objects():
    c0 = new_canvas(backend="matplotlib")
    c1 = new_canvas(backend="matplotlib")
    assert c0.title is not c1.title
    assert c0.x is not c1.x
    c0.title.text = "Title-0"
    c1.title.text = "Title-1"
    assert c0.title.text == "Title-0"
    assert c1.title.text == "Title-1"
    c0.x.color = "red"
    c1.x.color = "blue"
    assert_color_equal(c0.x.color, "red")
    assert_color_equal(c1.x.color, "blue")

def test_update_methods():
    canvas = new_canvas(backend="mock")
    canvas.update_axes(visible=False)
    canvas.update_labels(title="Title", x="X", y="Y")
    canvas.update_font(size=24, color="red", family="Arial")

def test_native_and_wrapping(backend: str):
    canvas = new_canvas(backend=backend)
    assert canvas.native is not None
    if backend == "mock":
        return
    new = wrap_canvas(canvas.native)
    repr(new._get_backend())
    assert new._get_backend() == canvas._get_backend()

def test_grid(backend: str):
    cgrid = wc.new_grid(2, 2, backend=backend).link_x().link_y()
    c00 = cgrid.add_canvas(0, 0)
    c01 = cgrid.add_canvas(0, 1)
    c10 = cgrid.add_canvas(1, 0)
    c11 = cgrid.add_canvas(1, 1)
    assert cgrid[0, 0] is not cgrid[1, 0]
    assert cgrid[0, 1] is not cgrid[1, 1]
    assert cgrid[0, 1].visible
    with filter_warning(backend, "plotly"):
        cgrid[0, 1].visible = False

    c00.add_line([0, 1, 2], [0, 1, 2])
    c01.add_hist([0, 1, 2, 3, 4, 3, 2, 1])
    c10.add_markers([3, 1, 2])
    c11.add_rug([1, 3, 2, 1.2])

    assert_allclose(c00.x.lim, c01.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.x.lim, c10.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.x.lim, c11.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.y.lim, c01.y.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.y.lim, c10.y.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.y.lim, c11.y.lim, rtol=1e-6, atol=1e-6)

    assert len(c00.layers) == 1
    assert len(c01.layers) == 1
    assert len(c10.layers) == 1
    assert len(c11.layers) == 1


def test_grid_nonuniform(backend: str):
    cgrid = wc.new_grid([2, 1], [2, 1], backend=backend, size=(100, 100)).link_x().link_y()
    repr(cgrid)
    c00 = cgrid.add_canvas(0, 0)
    c01 = cgrid.add_canvas(0, 1)
    c10 = cgrid.add_canvas(1, 0)
    c11 = cgrid.add_canvas(1, 1)

    c00.add_line([0, 1, 2], [0, 1, 2])
    c01.add_hist([0, 1, 2, 3, 4, 3, 2, 1])
    c10.add_markers([3, 1, 2])
    c11.add_rug([1, 3, 2, 1.2])

    assert_allclose(c00.x.lim, c01.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.x.lim, c10.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.x.lim, c11.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.y.lim, c01.y.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.y.lim, c10.y.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c00.y.lim, c11.y.lim, rtol=1e-6, atol=1e-6)

    assert len(c00.layers) == 1
    assert len(c01.layers) == 1
    assert len(c10.layers) == 1
    assert len(c11.layers) == 1

def test_vgrid_hgrid(backend: str):
    cgrid = wc.new_col(2, backend=backend, size=(100, 100)).link_x().link_y()
    repr(cgrid)
    c0 = cgrid.add_canvas(0)
    c1 = cgrid.add_canvas(1)
    assert cgrid[0] is not cgrid[1]

    c0.add_line([0, 1, 2], [0, 1, 2])
    c1.add_hist([0, 1, 2, 3, 4, 3, 2, 1])

    assert_allclose(c0.x.lim, c1.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c0.y.lim, c1.y.lim, rtol=1e-6, atol=1e-6)

    assert len(c0.layers) == 1
    assert len(c1.layers) == 1

    cgrid = wc.new_row(2, backend=backend, size=(100, 100)).link_x().link_y()
    c0 = cgrid.add_canvas(0)
    c1 = cgrid.add_canvas(1)
    assert cgrid[0] is not cgrid[1]

    c0.add_line([0, 1, 2], [0, 1, 2])
    c1.add_hist([0, 1, 2, 3, 4, 3, 2, 1])

    assert_allclose(c0.x.lim, c1.x.lim, rtol=1e-6, atol=1e-6)
    assert_allclose(c0.y.lim, c1.y.lim, rtol=1e-6, atol=1e-6)

    assert len(c0.layers) == 1
    assert len(c1.layers) == 1

def test_unlink(backend: str):
    grid = wc.new_row(2, backend=backend).fill()
    linker = wc.link_axes(grid[0].x, grid[1].x)
    grid[0].x.lim = (10, 11)
    assert grid[0].x.lim == pytest.approx((10, 11))
    assert grid[1].x.lim == pytest.approx((10, 11))
    linker.unlink_all()
    grid[0].x.lim = (20, 21)
    assert grid[0].x.lim == pytest.approx((20, 21))
    assert grid[1].x.lim == pytest.approx((10, 11))

    grid = wc.new_row(2, backend=backend).fill().link_x().link_y()

def test_jointgrid(backend: str):
    rng = np.random.default_rng(0)
    joint = wc.new_jointgrid(backend=backend, size=(100, 100)).with_hist().with_kde().with_rug()
    joint.add_markers(rng.random(100), rng.random(100), color="red")
    if backend != "vispy":
        joint.add_legend()

def test_legend(backend: str):
    if backend == "vispy":
        pytest.skip("vispy does not support legend")
    canvas = new_canvas(backend=backend)
    canvas.add_line([0, 1, 2], [0, 1, 2], name="line")
    canvas.add_markers([0, 1, 2], [0, 1, 2], name="markers")
    canvas.add_bars([0, 1, 2], [0, 1, 2], name="bars")
    canvas.add_line([3, 4, 5], [1, 2, 1], name="plot").with_markers()
    canvas.add_line([3, 4, 5], [1, 2, 1], name="plot").with_xband([1, 1, 1])
    canvas.add_line([3, 4, 5], [1, 2, 1], name="plot").with_yband([1, 1, 1])
    canvas.add_line([3, 4, 5], [2, 3, 2], name="line+err").with_yerr([1, 1, 1])
    canvas.add_markers([3, 4, 5], [3, 4, 3], name="markers+err").with_xerr([1, 1, 1])
    canvas.add_line([3, 4, 5], [4, 5, 4], name="plot+err").with_markers().with_xerr([1, 1, 1]).with_yerr([1, 1, 1])
    canvas.add_markers([3, 4, 5], [5, 6, 5], name="markers+err+err").with_stem()
    canvas.add_legend(location="bottom_right")

def test_animation():
    from whitecanvas.animation import Animation

    canvas = new_canvas(backend="matplotlib")
    anim = Animation(canvas)
    x = np.linspace(0, 2 * np.pi, 100)
    line = canvas.add_line(x, np.sin(x + 0), name="line")
    for i in anim.iter_range(3):
        line.set_data(x, np.sin(x + i * np.pi / 3))
    with tempfile.TemporaryDirectory() as tmpdir:
        anim.save(Path(tmpdir) / "test.gif")
    assert anim.asarray().ndim == 4

def test_layer_handling(backend: str):
    canvas = new_canvas(backend=backend)
    canvas.add_line([0, 1, 2], name="l_0")
    canvas.add_markers([0, 1, 2], name="l_1").with_xerr([0.1, 0.1, 0.1])
    repr(canvas.layers)
    assert canvas.layers[0].name == "l_0"
    assert canvas.layers[0] is canvas.layers["l_0"]
    assert canvas.layers.get("l_0") is canvas.layers[0]
    canvas.layers.move(1, 0)
    canvas.layers.pop()
    canvas.layers.pop()

def test_multidim():
    canvas = new_canvas(backend="matplotlib")
    x = np.arange(5)
    ys = [x, x ** 2, x ** 3]
    canvas.dims.in_axes("T").add_line(x, ys)
    canvas.dims.add_markers(x, ys, color="red", symbol="+")
    canvas.dims.add_rug(ys)
    canvas.dims.add_band(x, ys, [y0 + 1 for y0 in ys])
    canvas.dims.add_errorbars(x, ys, [y - 1 for y in ys])
    canvas.dims.add_text(x, ys, ["a", "b", "c", "d", "e"])
    img = np.zeros((2, 5, 5))
    canvas.dims.add_image(img)
    canvas.dims.set_indices(1)
    canvas.dims.set_indices({"T": 0})

def test_multidim_slider():
    # TODO: how to test other app backends?
    plt.close("all")
    canvas = new_canvas(backend="matplotlib:qt")
    x = np.arange(5)
    ys = [x, x ** 2, x ** 3]
    canvas.dims.add_line(x, ys)
    slider = canvas.dims.create_widget()
    slider._emit_changed()

def test_second_x(backend: str):
    if backend in ("vispy", "plotly"):
        pytest.skip(f"{backend} does not support second_x")
    canvas = new_canvas(backend=backend)
    other = canvas.install_second_x()
    canvas.add_line([0, 1, 2], [0, 1, 2], name="line")
    other.add_line([0, 1, 2], [0, 1, 2], name="line")
    canvas.y.lim = (0, 2)
    assert canvas.y.lim == pytest.approx((0, 2))
    assert other.y.lim == pytest.approx((0, 2))

def test_second_y(backend: str):
    if backend in ("vispy",):
        pytest.skip(f"{backend} does not support second_y")
    canvas = new_canvas(backend=backend)
    other = canvas.install_second_y()
    canvas.add_line([0, 1, 2], [0, 1, 2], name="line")
    other.add_line([0, 1, 2], [0, 1, 2], name="line")
    canvas.x.lim = (0, 2)
    assert canvas.x.lim == pytest.approx((0, 2))
    assert other.x.lim == pytest.approx((0, 2))

def test_inset(backend: str):
    if backend in ("vispy", "plotly", "bokeh"):
        pytest.skip(f"{backend} does not support inset")
    canvas = new_canvas(backend=backend)
    inset = canvas.install_inset((0.5, 0.96, 0.6, 0.96))
    canvas.add_line([0, 1, 2], [0, 1, 2], name="line")
    inset.add_line([0, 1, 2], [0, 1, 2], name="line")

def test_to_html(backend: str):
    if backend not in ("plotly", "bokeh"):
        pytest.skip(f"{backend} does not support to_html")
    canvas = new_canvas(backend=backend)
    canvas.add_line([0, 1, 2], [0, 1, 2], name="line")
    canvas.to_html()

def test_screenshot(backend: str):
    if backend in ("plotly", "bokeh", "mock"):
        pytest.skip(f"{backend} does not support screenshot")
    if backend == "vispy":
        backend += ":qt"
    canvas = new_canvas(backend=backend)
    canvas.add_line([0, 1, 2], [0, 1, 2], name="line")
    if sys.platform != "win32":  # NOTE: testing in GitHub Actions fails for some reason
        canvas._repr_png_()

def test_emulating_mouse_click():
    canvas = new_canvas("mock")

    clicked_history: list[Point] = []
    double_clicked_history: list[Point] = []

    @canvas.mouse.clicked.connect
    def _clicked(ev: MouseEvent):
        clicked_history.append(ev.pos)

    @canvas.mouse.double_clicked.connect
    def _double_clicked(ev: MouseEvent):
        double_clicked_history.append(ev.pos)

    canvas.mouse.emulate_click((1, 1))
    assert clicked_history == [(1, 1)]
    assert double_clicked_history == []

    canvas.mouse.emulate_click((1, 2))
    assert clicked_history == [(1, 1), (1, 2)]
    assert double_clicked_history == []

    canvas.mouse.emulate_double_click((-1, -1))
    assert clicked_history == [(1, 1), (1, 2)]
    assert double_clicked_history == [(-1, -1)]

def test_emulating_mouse_move():
    canvas = new_canvas("mock")
    history: list[str] = []

    @canvas.mouse.moved.connect
    def _move(ev: MouseEvent):
        if ev.button != "right":
            return
        history.append("press")
        yield
        while ev.type != "release":
            history.append("move")
            yield
        history.append("release")

    canvas.mouse.emulate_drag([(1, 1), (1, 2), (1, 3)], button="left")
    assert history == []
    canvas.mouse.emulate_drag([(1, 1), (1, 2), (1, 3)], button="right")
    assert history == ["press", "move", "move", "release"]

def test_canvas_3d(backend: str):
    if backend not in ("matplotlib", "vispy", "plotly"):
        pytest.skip(f"{backend} does not support 3d")
    canvas = new_canvas_3d(backend=backend)
    if backend != "vispy":
        canvas.update_axes(visible=True, color="gray")
        canvas.update_font(size=13, color="pink", family="Arial")
    canvas.update_labels(title="Title", x="X", y="Y", z="Z")
    canvas.add_line([0, 1, 2], [0, 1, 2], [0, 1, 2])
    canvas.add_markers([0, 1, 2], [0, 1, 2], [0, 1, 2])
    if backend == "matplotlib":
        canvas.add_vectors([0, 1, 2], [0, 1, 2], [0, 0, 0], [0, 1, 2], [0, 1, 2], [1, 1, 1])
    canvas.add_surface(np.array([[2,3,2], [3,2,3]]))
